import asyncio
from asyncio import Task
from contextlib import asynccontextmanager
from typing import Any, AsyncIterator

from redis import asyncio as aioredis

from redishilok.rwlock import RedisRWLock


class RedisRWLockCtx:
    def __init__(
        self,
        redis: str | aioredis.Redis,
        path: str,
        ttl: int = 10000,
        refresh_interval: float = 3000,
        cancel_on_lock_failure: bool = True,
    ) -> None:
        """Context manager for acquiring Redis read/write locks.

        Args:
            redis (str): Redis connection URL or aioredis.Redis connection.
            path (str): Key for the lock.
            ttl (int, optional): Lock TTL in milliseconds. Defaults to 10000.
            refresh_interval (int, optional): Lock refresh interval in milliseconds. Defaults to 3000.
            cancel_on_lock_failure (bool, optional): Cancel the current task if lock refresh fails. Defaults to True.

        Raises:
            RuntimeError: If lock refresh fails.
        """
        self._context_task: Task[Any] | None = None
        self.cancel_on_lock_failure = cancel_on_lock_failure
        self.refresh_interval = refresh_interval
        self.lock = RedisRWLock(redis, path, ttl)
        self._refresh_task: Task[Any] | None = None
        self._stop_event = asyncio.Event()

    @property
    def path(self) -> str:
        return self.lock.path

    @property
    def ttl(self) -> int:
        return self.lock.ttl

    async def close(self) -> None:
        """Close the redis connection."""
        await self._stop_refresh()
        await self.lock.close()

    async def _start_refresh(self, shared: bool) -> None:
        async def refresh_loop() -> None:
            try:
                while not self._stop_event.is_set():
                    try:
                        await self.lock.refresh_lock(shared=shared)
                    except RuntimeError as e:
                        # Lock lost; raise an exception to terminate the task
                        raise RuntimeError(f"Refresh failed: {str(e)}") from e
                    await asyncio.sleep(self.refresh_interval / 1000)
            except Exception:
                self._stop_event.set()
                if self._context_task and self.cancel_on_lock_failure:
                    self._context_task.cancel()
                raise

        self._refresh_task = asyncio.create_task(refresh_loop())

    async def _stop_refresh(self) -> None:
        if self._refresh_task:
            self._stop_event.set()
            self._refresh_task.cancel()
            try:
                await self._refresh_task
            except asyncio.CancelledError:
                pass
            self._refresh_task = None
            self._stop_event.clear()

    @asynccontextmanager
    async def read(
        self, block: bool = True, timeout: float | None = None
    ) -> AsyncIterator[None]:
        """Context manager for acquiring a read lock."""
        acquired = await self.lock.acquire_read_lock(block=block, timeout=timeout)
        if not acquired:
            raise RuntimeError(f"Failed to acquire read lock for {self.path}")
        try:
            await self._start_refresh(shared=True)
            yield
        finally:
            try:
                await self._stop_refresh()
            finally:
                await self.lock.release_read_lock()

    @asynccontextmanager
    async def write(
        self, block: bool = True, timeout: float | None = None
    ) -> AsyncIterator[None]:
        """Context manager for acquiring a write lock."""
        acquired = await self.lock.acquire_write_lock(block=block, timeout=timeout)
        if not acquired:
            raise RuntimeError(f"Failed to acquire write lock for {self.path}")
        try:
            self._context_task = asyncio.current_task()
            await self._start_refresh(shared=False)
            yield
        finally:
            try:
                await self._stop_refresh()
            finally:
                await self.lock.release_write_lock()



import asyncio
import os

from redis import asyncio as aioredis


class RedisRWLock:
    _redis: aioredis.Redis

    def __init__(
        self, redis_url_or_redis_conn: str | aioredis.Redis, path: str, ttl: int
    ):
        self._redis_param = redis_url_or_redis_conn
        self.path = path
        self.ttl = ttl
        self.uuid = os.urandom(16).hex()

    @property
    def redis(self) -> aioredis.Redis:
        if not hasattr(self, "_redis"):
            if isinstance(self._redis_param, str):
                self._redis = aioredis.from_url(self._redis_param)
            elif isinstance(self._redis_param, aioredis.Redis):
                self._redis = self._redis_param
            else:  # pragma: no cover
                raise ValueError(
                    "redis_url_or_redis_conn must be a string or aioredis.Redis instance."
                )
        return self._redis

    async def close(self) -> None:
        if self._redis is not None:
            await self.redis.aclose()

    async def acquire_read_lock(
        self, block: bool = True, timeout: float | None = None
    ) -> bool:
        script = """
        if redis.call("HGET", KEYS[1], "writer") ~= false then
            return false
        end
        redis.call("LPUSH", KEYS[2], ARGV[1])
        redis.call("PEXPIRE", KEYS[1], ARGV[2])
        redis.call("PEXPIRE", KEYS[2], ARGV[2])
        return true
        """
        readers_key = f"{self.path}:readers"
        while True:
            acquired = await self.redis.eval(  # type: ignore[misc]
                script, 2, self.path, readers_key, self.uuid, str(self.ttl)
            )
            if acquired or not block:
                return bool(acquired)
            if timeout is not None:
                timeout -= 0.1
                if timeout <= 0:
                    return False
            await asyncio.sleep(0.05)

    async def acquire_write_lock(
        self, block: bool = True, timeout: float | None = None
    ) -> bool:
        script = """
        if redis.call("LLEN", KEYS[2]) > 0 then
            return false
        end
        if redis.call("HGET", KEYS[1], "writer") ~= false then
            return false
        end
        redis.call("HSET", KEYS[1], "writer", ARGV[1])
        redis.call("PEXPIRE", KEYS[1], ARGV[2])
        return true
        """
        readers_key = f"{self.path}:readers"
        while True:
            acquired = await self.redis.eval(  # type: ignore[misc]
                script, 2, self.path, readers_key, self.uuid, str(self.ttl)
            )
            if acquired or not block:
                return bool(acquired)
            if timeout is not None:
                timeout -= 0.1
                if timeout <= 0:
                    return False
            await asyncio.sleep(0.1)

    async def refresh_lock(self, shared: bool = True) -> None:
        script = """
        if ARGV[1] == "shared" then
            if redis.call("LPOS", KEYS[2], ARGV[2]) == false then
                return false
            end
            redis.call("PEXPIRE", KEYS[1], ARGV[3])
            redis.call("PEXPIRE", KEYS[2], ARGV[3])
        else
            if redis.call("HGET", KEYS[1], "writer") ~= ARGV[2] then
                return false
            end
            redis.call("PEXPIRE", KEYS[1], ARGV[3])
        end
        return true
        """
        readers_key = f"{self.path}:readers"
        lock_type = "shared" if shared else "exclusive"
        refreshed = await self.redis.eval(  # type: ignore[misc]
            script, 2, self.path, readers_key, lock_type, self.uuid, str(self.ttl)
        )
        if not refreshed:
            raise RuntimeError(
                "Failed to refresh lock: Lock does not exist or is not held."
            )

    async def release_read_lock(self) -> bool:
        script = """
        local pos = redis.call("LPOS", KEYS[1], ARGV[1])
        if pos ~= false then
            redis.call("LREM", KEYS[1], 1, ARGV[1])
        end
        return true
        """
        readers_key = f"{self.path}:readers"
        return bool(await self.redis.eval(script, 1, readers_key, self.uuid))  # type: ignore[misc]

    async def release_write_lock(self) -> bool:
        script = """
        if redis.call("HGET", KEYS[1], "writer") == ARGV[1] then
            redis.call("HDEL", KEYS[1], "writer")
            return true
        end
        return false
        """
        return bool(await self.redis.eval(script, 1, self.path, self.uuid))  # type: ignore[misc]



import logging
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from typing import Any, AsyncIterator

from redis import asyncio as aioredis

from redishilok.rwctx import RedisRWLockCtx


class RedisHiLok:
    def __init__(
        self,
        redis: str | aioredis.Redis,
        ttl: int = 5000,
        refresh_interval: float = 2000,
        separator: str = "/",
        cancel_on_lock_failure: bool = True,
    ):
        if isinstance(redis, str):
            self.redis = aioredis.from_url(redis)
        else:
            self.redis = redis
        self.ttl = ttl
        self.refresh_interval = refresh_interval
        self.separator = separator
        self.cancel_on_lock_failure = cancel_on_lock_failure

    async def __aenter__(self) -> "RedisHiLok":
        return self

    async def __aexit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
        await self.close()

    async def close(self) -> None:
        await self.redis.aclose()

    def _build_lock(self, path: str) -> RedisRWLockCtx:
        return RedisRWLockCtx(
            self.redis,
            path,
            ttl=self.ttl,
            refresh_interval=self.refresh_interval,
            cancel_on_lock_failure=self.cancel_on_lock_failure,
        )

    async def _acquire_hierarchy(
        self, path: str, shared_last: bool, block: bool, timeout: float | None
    ) -> list[tuple[RedisRWLockCtx, AbstractAsyncContextManager[None]]]:
        nodes = list(filter(lambda x: x, path.split(self.separator)))
        locks: list[tuple[RedisRWLockCtx, AbstractAsyncContextManager[None]]] = []
        try:
            for i, node in enumerate(nodes):
                lock_path = self.separator.join(nodes[: i + 1])
                lock = self._build_lock(lock_path)
                if i < len(nodes) - 1:  # Ancestors: always shared
                    lock_ctx = lock.read(block=block, timeout=timeout)
                else:  # Target node: mode depends on `shared_last`
                    if shared_last:
                        lock_ctx = lock.read(block=block, timeout=timeout)
                    else:
                        lock_ctx = lock.write(block=block, timeout=timeout)
                await lock_ctx.__aenter__()
                locks.append((lock, lock_ctx))
            return locks
        except Exception:
            await self._release_hierarchy(locks)
            raise

    @staticmethod
    async def _release_hierarchy(
        locks: list[tuple[RedisRWLockCtx, AbstractAsyncContextManager[None]]],
    ) -> None:
        for i, (lock, ctx) in enumerate(reversed(locks)):
            try:
                await ctx.__aexit__(None, None, None)
                await lock.close()
            except Exception:
                # this isn't catastrophic, but we should log it
                logging.exception("Failed to release hilok")

    @asynccontextmanager
    async def read(
        self, path: str, block: bool = True, timeout: float | None = None
    ) -> AsyncIterator[None]:
        locks = await self._acquire_hierarchy(
            path, shared_last=True, block=block, timeout=timeout
        )
        try:
            yield
        finally:
            await self._release_hierarchy(locks)

    @asynccontextmanager
    async def write(
        self, path: str, block: bool = True, timeout: float | None = None
    ) -> AsyncIterator[None]:
        locks = await self._acquire_hierarchy(
            path, shared_last=False, block=block, timeout=timeout
        )
        try:
            yield
        finally:
            await self._release_hierarchy(locks)
